#include <math.h>
#include <stdlib.h>
/* #include <GL/glut.h> */
#include <string.h>

#include "../src/grind.h"
#include "graphics.h"

static _grcall clear_grcall = {CLEAR,NULL};
_grcall *first_grcall = &clear_grcall;
_grcall *last_grcall = &clear_grcall;

extern const struct timespec delay;

int addGrcall(_grtype t, void *dat, double nx, double ny, double nz) {
  _grcall *tmp_grcall;
  int *col;
  char *str;
  double *coord;

  /*  check whether the stack has been initialized */
  if (first_grcall == NULL) {
    fprintf(stderr,"error: adding to graphics stack while first element is empty\nfirst element should be \"primitive 'clear'\"\n\n");
    exit (0);
  }

  /*  allocate memory for new stack element */
  if (!(tmp_grcall=(_grcall*)malloc(sizeof( _grcall)))) {
    fprintf(stderr, "error allocating mem in addGrcall: tmp_grcall\n");
    exit(1);
  }
  /*  set it's values */
  tmp_grcall->type = t;
  tmp_grcall->next = NULL;
  /*  handle some special cases */
  if (t == TEXT) {
    str = ccalloc(strlen((char*)dat)+1, "'str' in addGrcall()");
    strcpy(str, (char*)dat);
    tmp_grcall->val = str;
  }
  else if (t == RTEXT) {
    str = ccalloc(strlen((char*)dat)+1, "'str' in addGrcall()");
    strcpy(str, (char*)dat);
    tmp_grcall->val = str;
  }
  else if (t == LTEXT) {
    str = ccalloc(strlen((char*)dat)+1, "'str' in addGrcall()");
    strcpy(str, (char*)dat);
    tmp_grcall->val = str;
  }
  else if (t == COLOR) {
    col = icalloc(1, "'col' in addGrcall()");
    *col = *(int*)dat;
    tmp_grcall->val = col;
  }
  else if (t == MOVE || t == DRAW || t == GR_POINT || t==ORECT || t==FRECT) {
    coord = dcalloc(3, "'dbl' in addGrcall()");
    coord[0] = nx; coord[1] = ny; coord[2] = nz;
    tmp_grcall->val = coord;
  }
  else
    tmp_grcall->val = NULL;

  /*  lock the stack so that other threads reading the stack can't interfere */
  append2stack(&last_grcall, tmp_grcall);
  return 1;
}

int DeleteCall(_grcall *tmp_grcall) {
  int t = tmp_grcall->type;
  /*  handle the special cases */
  if (t == TEXT || t == LTEXT || t == RTEXT || 
      t == COLOR || t == MOVE || t == DRAW || 
      t == GR_POINT || t==ORECT || t==FRECT)
    free(tmp_grcall->val);
  free(tmp_grcall);
  return 1;
}

int eraseGRstack() {
  /*  erase the stack starting from the second element; */
  /*  the first element (ie the 'clear' command) remains as first element */
  _grcall *next_grcall = first_grcall->next;
  _grcall *del_grcall;

  /*  firstly, set 'last' pointer to first element of the stack */
  /*  if we get an interrupt during erase at least the global pointers are ok */
  last_grcall = first_grcall;
  /*   and de-couple the stack from the rest which is to be erased */
  first_grcall->next = NULL;

  while (next_grcall != NULL) {
    del_grcall = next_grcall;
    next_grcall = next_grcall->next;
    del_grcall->next = NULL;
    DeleteCall(del_grcall);
  }
  return 1;
}
