#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "grind.h"
#include "list.h"

extern FILE *cinput;
extern char uline[];
extern char cline[];
extern char *cbegin;
extern char *cend;
extern char *modelname;

extern double finish;
extern int nintervals;
extern double *eigvector;
extern int  *eigsort, *eigstab, *eigcomp;

extern int mark_opt;
extern _grcall *first_grcall;
extern int itor;

extern boolean forward_opt;
extern const char *itors[];

static Node *macrolist = NULL;

#define n_commands 51
static const _symb_table command_table[n_commands] = {
  {"2d",            TwoD},
  {"3d",            ThreeD},
  {"axis",          Axis},
  {"calculator",    Calc},
  {"bye",           Bye},
  {"quit",          Bye},
  {"end",           Bye},
  {"stop",          Bye},
  {"bifurcate",     Bifurcate},
  {"continue",      Continue},
  {"cursor",        Cursor},
  {"display",       Display},
  {"e2n",           E2n},
  {"e2r",           E2r},
  {"e3n",           E3n},
  {"e3r",           E3r},
  {"eigen",         Eigen},
  {"erase",         Erase},
  {"ert",           Ert},
  {"finish",        Finish},
  {"format",        Format},
  /* {"graphics",      Graphics}, */
  {"grid",          Grid},
  {"integrator",    Integrator},
  {"help",          Help},
  {"keepvar",       Keepvar},
  {"load",          Load},
  {"macro",         Macro},
  {"mine",          mine},
  {"mode",          Mode},
  {"newton",        Newton},
  {"nep",           Nep},
  {"noise",         Noise},
  {"nullcline",     Nullcline},
  {"isocline",      Nullcline},
  {"option",        Option},
  {"output",        Output},
  {"parameters",     Parameter},
  {"perturb",       Perturb},
  {"plot",          Plot},
  {"poincare",      Poincare},
  {"export",        Export},
  {"readfile",      Readfile},
  {"rgb",           GR_RGB},
  {"run",           Run},
  {"shade",         Shade},
  {"takens",        Takens},
  {"timeplot",      Timeplot},
  {"triangle",      Triangle},
  {"value",         Value},
  {"vector",        Vector},
  {"where",         Where}
};

int Assign(int argc,char *argv[]) {
  /*  command 'assign':  */
  /*  first argument is assumed to be the name of a variable or a parameter */
  /*  the second argument is an expression which evaluates to a (real) value, */
  /*  as calculated by 'calc()' */
  int i;
  double v;

  extern char *cbegin;
  
  if ((i = nameIndex(argv[0])) < 0 || i >= nnames)
    return gr_warning("assignment to unknown argument: %s\n", argv[0]);
  if (argc < 3)
    return 0;

  /*  first read the second argument properly from uline */
  v = calc(strstr(strstr(cbegin,"="), argv[2]));
  switch(all_types[i]) {
  case VARIABLE:
    istate[i] = v;
    vars[i].init = true;
    break;
  case EXPRESSION:
    all_values[i] = v;
    exprs[i-nvar].init = true;
    break;
  case PARAMETER:
    pars[i-nvar-nexpr].value = all_values[i] = v;
    pars[i-nvar-nexpr].init = true;
    break;
  default:
    gr_warning("assignment to unknown argument: %s\n", cbegin);
    break;
  }
  return 1;
}

int Axis(int argc,char *argv[]) {
  /*  routine to set axis parameters */
  int axi;
  int vari, ind;
 
  if (eqArg(1, "help", argc, argv)) {
    gHelp("axis x/y/z/t variable minimum maximum nticks [log]","axis x x 0 1 3",
	  "Defines an axis of a phase space (or Takens space)");
    gHelp("axis h/v minimum maximum nticks [log]","axis h 0 finish 4; axis v 0 1 3",
	  "Defines an axis of a time plot");
    return 1;
  }
  if (eqArg(1, "status", argc,argv)) {
    printf("axis\t\tvar/par\tminimum\tmaximum\tnticks\tlog\n\n");
    printf("%-15s %-7s %-7.1f %-7.1f %-7d %-7d\n",
	   "x-axis", axis[0].label, axis[0].min, axis[0].max,
	   axis[0].ticks, axis[0].log);
    printf("%-15s %-7s %-7.1f %-7.1f %-7d %-7d\n",
	   "y-axis", axis[1].label, axis[1].min, axis[1].max,
	   axis[1].ticks, axis[1].log);
    printf("%-15s %-7s %-7.1f %-7.1f %-7d %-7d\n",
	   "z-axis", axis[2].label, axis[2].min, axis[2].max,
	   axis[2].ticks, axis[2].log);
    printf("%-15s %-7s %-7.1f %-7.1f %-7d %-7d\n",
	   "takens-axis", axis[3].label, axis[3].min, axis[3].max,
	   axis[3].ticks, axis[3].log);
    printf("%-15s %-7s %-7.1f %-7.1f %-7d %-7d\n",
	   "vert. time-axis", axis[4].label, axis[4].min, axis[4].max,
	   axis[4].ticks, axis[4].log);
    printf("%-15s %-7s %-7.1f %-7.1f %-7d %-7d\n",
	   "hor. time-axis", axis[5].label, axis[5].min, axis[5].max,
	   axis[5].ticks, axis[5].log);
    return 1;
  }

  if (1 == argc)
    return 0;
  axi = (argc>1 ? Index("xyztvh",argv[1]) : -1);
  if (axi < 0)
    return gr_warning("not an axis: %s\n", argv[1]);

  /*  last arg may specify 'log'; set it now and then we can return anytime */
  if (abbrevArg(argc-1, "log", argc, argv)) {
    axis[axi].log = true;
    argc--;
  }
  else
    axis[axi].log = false;

  /*  we read 2nd argument */
  ind = 2;
  /*  if X, Y, Z, or TAKENS axis */
  if (axi <= 3) {
    /*  2nd arg is var-name or min-value */
    /*  this means that var-name is now optional also for X, Y, Z, TAKENS-axis */
    if (argc > ind) {
      vari = nameIndex(argv[ind]);
      if (vari >= 0 && vari < nnames) {
	axis[axi].var = vari;
	axis[axi].label = storeString(axis[axi].label,argv[ind]);
	ind++;
      }
    }
  }
  if (ind >= argc)
    /*  return if we're done at this point */
    return 0;
  /*  argv[ind] should be the min-value, but first check for typo's */
  if (isnumeric(ind, argc, argv))
    axis[axi].min = atof(argv[ind++]);
  else 
    return gr_warning("read error\n");
  /*  this arg should be the max-value */
  if (isnumeric(ind, argc, argv))
    axis[axi].max = atof(argv[ind++]);
  else 
    return gr_warning("read error\n");
  /*  if this arg exists and is numerical then it is the no. of ticks */
  if (argc>ind) {
    if (isnumeric(ind, argc, argv))
      axis[axi].ticks = atoi(argv[ind++]);
    else 
      return gr_warning("read error\n");
  }
  return 1;
}

int Bye(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("Bye"," ","Quit from GRIND");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for bye\n");

  if (!eqArg(0, "bye", argc, argv) && 
      !eqArg(0, "quit", argc, argv) && 
      !eqArg(0, "stop", argc, argv))
    return gr_warning("command not recognized: %s\n", argv[0]);

  closegraphics();
  exit (0);
  return 1;
}

int Calc(int argc, char *argv[]) {
  extern char *cbegin;

  if (eqArg(1, "help", argc, argv))
    return 
      gHelp("calculator expression", " ", "Calculator evaluates a formula");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for calc\n");

  if (argc<2)
    return 0;
  printf(dformat,calc(strstr(cbegin+strlen(argv[0]), argv[1])));
  printf("\n");
  return 1;
}

int Continue(int argc, char *argv[]) {
  static double *yt = NULL;
  static int cont_ax = 0;
  static double cont_step = 0.01;
  int i;
  if (NULL == yt) yt = dcalloc(nvar, "'yt' in Continue()");

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("continue x/y/z arclength","continue x 0.01",
		 "Continue an equilibrium along an parameter-axis");
  if (eqArg(1, "status", argc, argv))
    return printf("Status continue: axis=%s arclength=%.3g\n", 
		  axis[cont_ax].label, cont_step);

  if (argc > 1) {
    i = Index("xyz", argv[1]);
    if (i<X_AXIS || i>Z_AXIS)
      return gr_warning("not an axis: %s\n", argv[1]);
    cont_ax = i;
  }
  if (argc>2) {
    if (isnumeric(2, argc, argv))
      cont_step = atof(argv[2]);
    else
      return gr_warning("read error\n");
  }
  coparr(nvar,yt,nstate);
  if (!contEqui(cont_ax, cont_step,  1)) 
    return 0;
  coparr(nvar,nstate,yt); 
  return contEqui(cont_ax, cont_step, -1);
}

int Cursor(int argc,char *argv[]) {
  int i;
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("cursor"," ","Clicking in the graphics window sets initial state");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for cursor\n");

  if (mode_opt != DIM2)
    return gr_warning("this is not a 2D state-space\n");
  if (!mouse2state(istate))
    return gr_warning("error while reading mouse coordinates\n");
  for (i=0; i<nvar; i++)
    vars[i].init = true;
  return 1;
}

int Display(int argc,char *argv[]) {
  FILE *fp;
  static char *filename = NULL;
  char *patt = NULL;
  static boolean first = true;

  if (first) {
    first = false;
    filename = storeString(filename, modelname);
    if (!(fp = fopen(filename,"r"))) {
      filename = (char*)realloc(filename,(strlen(filename)+5)*sizeof(char));
      filename = strcat(filename, ".grd");
    }
  }

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("display filename [pattern]", "display model", 
		 "Copy file contents to screen");
  if (eqArg(1, "status", argc, argv))
    return printf("Status display: filename=%s, [pattern=%s]\n",
		  filename, (patt?patt:""));

  /*  opening the file for displaying */
  if (argc > 1)
    filename = storeString(filename, argv[1]);

  if ( !(fp = fopen(filename,"r")) )
    return gr_warning("unable to open file %s\n", filename);

  /*  check if a pattern is given to "grep" the file */
  if (argc > 2)
    patt = storeString(patt, argv[2]);

  /*  grep and display the file */
  grepfile(fp, patt);

  /*  close and free */
  fclose(fp);
  if (patt)
    free (patt);
  return 1;
}

int E2n(int argc,char *argv[]) {  
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("e2n"," ","Executes erase;2d;nullcline");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for '%s'\n", argv[0]);
  Erase(argc, argv);
  TwoD(argc,argv);
  Nullcline(argc,argv);
  return 1;
}

int E2r(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("e2r"," ","Executes erase;2d;run");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for '%s'\n", argv[0]);
  Erase(argc, argv);
  TwoD(argc,argv);
  Run(argc,argv);
  return 1;
}

int E3n(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("e3n"," ","Executes erase;3d;nullcline");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for '%s'\n", argv[0]);
  Erase(argc, argv);
  ThreeD(argc,argv);
  Nullcline(argc,argv);
  return 1;
}

int E3r(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("e3r"," ","Executes erase;3d;run");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for '%s'\n", argv[0]);
  Erase(argc, argv);
  ThreeD(argc,argv);
  Run(argc,argv);
  return 1;
}

int Eigen(int argc,char *argv[]) {
  /*  perform neighbourhood stability analysis, print eigenvalues and  */
  /*  eigenvectors of the jacobian in the current final state */
  static double *yt = NULL; /*  working state-vector */
  boolean stability, havejac = false;

  if ( NULL == yt ) yt = dcalloc(nvar, "'yt' in Eigen()");

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("eigen [jac]", " ", "Calculate eigenvalues and eigenvectors");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for eigen\n");

  if (2 == argc) {
    if (!(havejac = eqArg(1,"jac", argc, argv)))
      return gr_warning("read error: %s\n", argv[1]);
  }
  else if (argc>2)
    return gr_warning("reading too many arguments\n");

  /*  check whether jacobian exists */
  if (!jacob_opt)
    return gr_warning("No Jacobian, no eigenvalues\n");

  /*  copy the final (=newton?) state to a working state-vector */
  coparr(nvar, yt, fstate);
  /*  compute stability, eigen-values, and -vectors */
  stability = eigen(yt,1);
  if (stability == NEUTRAL)
    printf("Bifurcation point:\n");
  else if (stability == STABLE)
    printf("Stable point:\n");
  else
    printf("Unstable point:\n");
  /*  print derivatives */
  prider(yt);
  /*  print eigen-values and -vectors */
  prieig(yt, eqArg(argc-1,"jac", argc, argv) );
  if(mark_opt && (mode_opt >= DIM2)) 
    setMark(yt, (stability == STABLE));
  return 1;
}

int Erase(int argc, char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("erase"," ","Clears the graphics screen");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for erase\n");
  
  Signal2ResetStack();
  mode_opt = mode_none;
  return 1;
}

int Ert(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("ert"," ","Executes erase;run;timeplot");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for '%s'\n", argv[0]);

  Erase(argc, argv);
  Run(argc, argv);
  Timeplot(argc, argv);
  return 1;
}

int Export(int argc,char *argv[]) {
  static char *filename = NULL;

  if (NULL == filename) {
    filename = ccalloc(strlen(modelname)+5,"filename in export");
    strcpy(filename,modelname);
    strcat(filename,".eps");
  }
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv))
    return gHelp("export file","export grind",
          "Exports the current graphics window into an epsfile");
  if (eqArg(1, "status", argc, argv))
    return printf("Status export: file=%s\n", filename);

  if (!checkgraphics())
    return gr_warning("Not in graphical mode\n");

  export(argc, argv, &filename);

  return 1;
}

int Finish(int argc,char *argv[]) {
  /*  set new values for the variables finish (gives duration of a run) */
  /*  check first whether help or status is asked for */
  double f;
  int i;
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("finish time nintervals","finish 100.0 10",
		 "Set the integration time and report interval");
  if (eqArg(1, "status", argc, argv))
    return printf("Status finish: time = %-.1e; nintervals = %d\n",
		  finish, nintervals);

  if (argc>1) {
    if (isnumeric(1, argc, argv)) {
      f = atof(argv[1]);
      if (TINY < f) {
	finish = f;
	/*  adjust the horizontal axis of timeplots */
	/*  'finish' should be the new max-value */
	axis[H_AXIS].max = finish;
      }
      else 
	return gr_warning("finish %s too small\n",argv[1]);
    }
    else
      return gr_warning("read error\n");
  }
  if (argc>2) {
    if (isnumeric(2, argc, argv)) {
      i = (int)(atof(argv[2]));
      if (i) 
	nintervals = i;
      else
	return gr_warning("nintervals %s too small\n",argv[2]);
    }
    else
      return gr_warning("read error\n");
  }
  return 1;
}

int Format(int argc,char *argv[]) {
  /*  routine to set the format which is used to print doubles */
  
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("format print/eigen/vector/axis ndigits",
		 "format - 4 ","Define the format of numbers using C code");
  if (eqArg(1, "status", argc, argv)) {
    printf("Status of formats:\n");
    printf(sformat, "print:");
    printf("\"%s\"\n", dformat);
    printf(sformat, "eigen:");
    printf("\"%s\"\n", lformat);
    printf(sformat, "vector:");
    printf("\"%s\"\n", vformat);
    printf(sformat, "axis:");
    printf("\"%s\"\n", aformat);
    return 1;
  }

  if (argc == 1)
    return 0;

  if (eqf(argv[1], "print") && argc>2 && isdigit((int)argv[2][0])) {
    dformat[8] = argv[2][0];
    return 1;
  }
  if (eqf(argv[1], "eigen") && argc>2 && isdigit((int)argv[2][0])) {
    lformat[8] = argv[2][0];
    return 1;
  }
  if (eqf(argv[1], "vector") && argc>2 && isdigit((int)argv[2][0])) {
    vformat[8] = argv[2][0];
    return 1;
  }
  if (eqf(argv[1], "axis") && argc>2 && isdigit((int)argv[2][0])) {
    aformat[6] = argv[2][0];
    return 1;
  }
  if (isdigit(argv[1][0]))
    dformat[8] = argv[1][0];
  return 1;
}

int Graphics(int argc, char *argv[]) {  
  if (mode_opt<DIM2)
    mode_opt = DIM2;
  startgraphics(first_grcall);
  return 1;
}

int Grid(int argc, char *argv[]) {
  static int nprun = 5;
  double stepsize, zz, x, y;
  int i, j;
  boolean zstate = true;
  static double *dpop = NULL;

  if (eqArg(1, "help", argc, argv)) 
    return gHelp("grind npoints","grid 5",
		 "Start trajectories from npoints x npoints in phase space");
  if (eqArg(1, "status", argc, argv)) 
    return printf("Status grid: npoints=%d\n", nprun);

  if (NULL == dpop) dpop = dcalloc(nvar, "'dpop' in Load()");

  chckin();
  if (backward_opt) {
    forward_opt = false;
    printf("Derivatives run backwards!\n");
  }
  if (argc>1) {
    if (isnumeric(1, argc, argv))
      nprun = atoi(argv[1]);
    else
      return gr_warning("read error\n");
  }
  /*  distance between grid points */
  stepsize = (axis_opt ? 1./((double)nprun-1.0) : 1./(double)nprun);
  /*  start on top of axis or not   */
  zz = (axis_opt ? 0.0 : stepsize/2.);
  storeaxis(true);
  for (i=0, x=zz; i<nprun; i++, x+=stepsize) {
    setax(istate, x, X_AXIS);
    for (j=0, y=zz; j<nprun; j++, y+=stepsize) {
      setax(istate, y, Y_AXIS);
      coparr(nvar, fstate, istate);
      if (zstate) {
	fcn(istate, dpop);
	if (initial_opt) 
	  triangle(istate, dpop);
      }
      integr(NULL);
    }
  }
  storeaxis(false);
  printf(" End of %s integrations\n",itors[itor]);
  forward_opt = true;

  return 1;
}

int Help(int argc,char *argv[]) {
  int i, n;
  const int nrow = n_commands/3 + 1;
  static char *ask4help[] = {NULL, NULL};
  boolean print_all = false;

  if (NULL == ask4help[1])
    ask4help[1] = storeString(ask4help[1], "help");

  /*  check first whether help or status is asked for */
  if (eqArg(1, "status", argc, argv))
    return printf("No status for help\n");
  if (eqArg(1, "help", argc, argv)) {
    printf("Use: 'command help' to get help\n");
    print_all = true;
  }
  if (print_all || argc == 1 || eqArg(1, "all", argc, argv)) {
    printf("GRIND knows about the following commands:\n");
    for (i=0; i<nrow; i++) {
      printf(sformat, command_table[i].name);
      printf("\t\t");
      if (i+nrow < n_commands)
	printf(sformat, command_table[i+nrow].name);
      printf("\t\t");
      if (i+2*nrow < n_commands)
	printf(sformat, command_table[i+2*nrow].name);
      printf("\n");
    }
    return 1;
  }

  for (i=1; i<argc; i++) {
    /*  for each remaining argument call the corresponding command for help */
    for (n=0; n<n_commands; n++)
      if (eqf(argv[i], command_table[n].name)) {
	/*  execute the command with argument "help" */
	ask4help[0] = storeStr(command_table[n].name);
	command_table[n].funcptr(2, ask4help);
	free(ask4help[0]);
	break;
      }
    if (n_commands == n)
      gr_warning("not recognized: %s\n", argv[i]);
  }
  return 1;
}

int Integrator(int argc,char *argv[]) {
  extern double delta;
  extern double ztol;

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("integrator row4a/rk/euler tolerance/step","row4a 1e-6",
		 "Select the integrator");
  if (eqArg(1, "status", argc, argv))
    return printf("Status integrator: %s tolerance = %.3g, step = %.3g\n",
		  itors[itor], ztol, delta);

  if (argc == 1)
    return 0;

  if (eqf(argv[1], "row4a") ||
      eqf(argv[1], "ROW4a") ||
      eqf(argv[1], "ROW4A")) {
    itor = ROW4A;
    if (argc>2)
      ztol = atof(argv[2]);
  }
  if (eqf(argv[1], "euler") ||
      eqf(argv[1], "EULER")) {
    itor = EULER;
    if (argc>2)
      delta = atof(argv[2]);
  }
  if (eqf(argv[1], "rk") ||
      eqf(argv[1], "RK")) {
    itor = RK;
    if (argc>2)
       ztol = atof(argv[2]);
  }
  return 1;
}  

int Keepvar(int argc,char *argv[]) {
  /*  copy final state to initial state */
  int i;
  extern int ngrun; /* counter for no. of graphical runs; determines color */

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("keepvar"," ",
		 "Copies the final state into the initial state");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for keep\n");
      
  for (i=0; i<nvar; i++) {
    istate[i] = fstate[i];
  }
  ngrun--;
  return 1;
}

int checkRunBuffer(int); /* integ.c */
int Load(int argc,char *argv[]) {
  FILE *datafile;
  char *filename = NULL;
  static boolean first = true;
  boolean *orig_outputbuf = outputbuf;
  int nstep, i, n;
  boolean readerror = false;
  double fbuf;
  static double *pop = NULL, *dpop = NULL;
  int c;

  if (first) {
    /*  temp vectors holding state and derivative vectors */
    pop = dcalloc(nvar, "'pop' in Load()");
    dpop = dcalloc(nvar, "'dpop' in Load()");
  }

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("load filename"," ",
		 "Reads a file with time+nvar columns as if this was a run");
  if (eqArg(1, "status", argc, argv))
    return printf("Status load: filename=%s\n", filename);

  /*  opening the file for loading */
  if (argc>1)
    filename = storeString(filename, argv[1]);
  if ((datafile = fopen(filename,"r")) == NULL)
    return gr_warning("unable to open file %s\n", filename);

  /* temporarily resetting outputbuf */
  outputbuf = bcalloc(nvar, "'outputbuf' in Load()");
  /*  by default data is read only for all var's  */
  /*  and in addition calculated and written for expr's */
  for (i=0; i<nvar+nexpr; i++)
    outputbuf[i] = true;

  /*  find out number of lines in the datafile;  */
  /*  we need this in order to allocate  a large enough buffer */
  nstep = 0;
  while ((c = fgetc(datafile)) != EOF && c != '\n') {
    if ('\n' == c)
      nstep++;
  }
  rewind(datafile);
  /*  check and possiby allocate databuffer */
  checkRunBuffer(nstep);
  for (i=0; i<nstep; i++) {
    /*  read time */
    if (1 != fscanf(datafile, "%lf", &fbuf)) {
      readerror = true;
      /*  nstep holds number of steps read */
      nstep = i;
      break;
    }
    rundata[0][i] = (double)fbuf;
    /*  read all variable values; store in 'pop[]' */
    for (n=0; n<nvar; n++) {
      if (1 != fscanf(datafile, "%lf", &fbuf)) {
	readerror = true;
	/*  nstep holds number of steps read */
	nstep = i;
	break;
      }
      pop[n] = (double)fbuf;
    }
    /*  calculate derivatives (and thus set the expression values) */
    fcn(pop, dpop);
    /*  copy values to run-buffer */
    for (n=0; n<nvar+nexpr; n++)
      rundata[n+1][i] = all_values[n];
  }
  if (readerror) {
      gr_warning("error while reading datafrom file %s\n", filename);
    if (nstep <= 0) {
      finish = 0.;
      nintervals = 1;
      gr_warning("returning without having read any data\n");
      return 0;
    }
  }
  nintervals = nstep-1;
  finish = rundata[0][nintervals];
  printf("Succesfully read %d points", nstep);
  if (outputbuf)
    free(outputbuf);
  outputbuf = orig_outputbuf;
  fclose(datafile);
  if (filename)
    free(filename);
  return 0;
}

int Macro(int argc,char *argv[]) {
  char *cptr;
  extern char *cbegin;
  extern char *cend;

  if(argc == 1) {
    PrintList(macrolist);
    return 1;
  }

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("macro name command;command"," ",
		 "Define a macro with a number of assigments and/or commands");
  if (eqArg(1, "status", argc, argv)) {
    printf("The following macros are defined:\n");
    PrintList(macrolist);
  }

  if (argc == 2) {
    macrolist = DeleteNode(macrolist, argv[1]);
    return 1;
  }

  if (argc > 2) {
    cptr = strstr(strstr(cbegin," "), argv[1]);
    /*  find the beginning of the new macro in uline; it starts at 'argv[2]' */
    if ( (cptr = strstr(strstr(cptr," "), argv[2])) != NULL ) {
      /*  insert new macro in list */
      macrolist = InsertNode(macrolist, argv[1], cptr);
      /*  let cend point to the end of 'uline'  */
      /*  to indicate we're finished with current user input line */
      while(*(++cend) != '\0')
	;
    }
    else
      gr_warning("error defining new macro: %s\n", argv[2]);
    return 1;
  }

  return 1;
}

int Mode(int argc,char *argv[]) {
  /*  int t; */

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("mode alpha/2d/3d/none","mode alpha",
	  "Select alpha-numerical or graphical output mode");
  if (eqArg(1, "status", argc, argv)) {
    printf("current report mode = %s\n", 
	   ((mode_opt == 0) ? "none" :
	    ((mode_opt == 1) ? "alpha-numeric" :
	     ((mode_opt == 2) ? "2d-graphics" :
	      ((mode_opt == 3) ? "3d_graphics" : "unknown mode")))));
    return 1;
  }

  /*  read and interpret arguments */
  if (1 == argc) {
    mode_opt = mode_none = ALPHA;
    return 1;
  }
  if (abbrevArg(1, "alpha", argc, argv))
    mode_opt = mode_none = ALPHA;
  else if (abbrevArg(1, "2d", argc, argv)) {
    if (!checkgraphics())
      Graphics(1, NULL);
    mode_opt = DIM2;
  }
  else if (abbrevArg(1, "3d", argc, argv)) {
    if (!checkgraphics())
      Graphics(1, NULL);
    mode_opt = DIM3;
  }
  else if (abbrevArg(1, "none", argc, argv))
    mode_opt = mode_none = NONE;
  else {
    printf("usage: mode [alpha/2d/3d/none]\n");
    return 0;
  }
  return 1;
}

int Nep(int argc,char *argv[]) {
  
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("nep"," ","Execute newton;eigen;perturb");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for '%s'\n", argv[0]);

  Newton(argc,argv);
  Eigen(argc,argv);
  Perturb(argc, argv);
  return 1;
}

double tolnew = 1.e-6; /*  also used in 'continue' */
int Newton(int argc,char *argv[]) {
  /*  search an equilibrium using Newton-Rhapson's method */
  /*  as second argument the tolerance can be given (in the future) */
  int i;
  int newtrls;

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("newton tolerance","newton 1e-9",
		 "Approach an equilibrium by a Newton Raphson iteration");
  if (eqArg(1, "status", argc, argv))
      return printf("Status newton: %.3g\n", tolnew);

  /*  check whether everything is initialized */
  chckin();
  /*  copy iniial state to final state; work will be done on the latter */
  for (i=0; i<nvar; i++)
    fstate[i] = istate[i];
  /*  check for extra argument which specifies a new tolerance value */
  if (argc>1) {
    if (isnumeric(1, argc, argv))
      tolnew =  atof(argv[1]);
    else
      return gr_warning("read error\n");
  }
  /*  no. of trials in mnewt */
  newtrls = 25;
  /*  from NR */
  if ( 0 == mnewt(newtrls, fstate-1, nvar, tolnew, tolnew) ){
    printf("No convergence\n");
    prider(fstate);
  }
  else {
    printf("Converged into:\n");
    prider(fstate);
  }
  for (i=0; i<nvar; i++)
    nstate[i] = fstate[i];
  return 1;
}

int Noise(int argc, char *argv[]) {
  int i, p = 0;
  double d;

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("noise parameter mean stdev [pos]"," ",
		 "Set Gaussian noise on a parameter");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for noise\n");

  if (abbrevArg(1, "off", argc, argv)) {
    for (i=0; i<npar; i++) {
      pars[i].noise = false;
      pars[i].pos = false;
    }
    return 1;
  }

  /*  remaining arguments is a parameter for which a value and stdev is set */
  /* for (i=1; i<argc; i++) { */
  if (0 == argc)
    return 0;
  i = 1;
  if ( (p = nameIndex(argv[i])) >= nvar+nexpr && p < nnames) {
    p -= (nvar + nexpr);
  }
  else
    return gr_warning("not a parameter: %s\n", argv[1]);
  i++;
  if (eqArg(i, "off", argc, argv)) {
    pars[p].pos = pars[p].noise = false;
    pars[p].sd = 0.0;
    return 1;
  }
  /*  else ... */
  pars[p].init = true;
  if (argc>i) {
    if (isnumeric(i, argc, argv))
      par_values[p+nexpr] = pars[p].value = atof(argv[i]);
    else
      return gr_warning("read error: expected a numerical argument: %s\n", 
			argv[i]);
  }
  else
    return 1;
  i++;
  if (eqArg(i, "off", argc, argv)) {
    pars[p].pos = pars[p].noise = false;
    pars[p].sd = 0.0;
    return 1;
  }
  if (argc>i) {
    if (isnumeric(i, argc, argv)) {
      d = atof(argv[i]);
      if (d<0.)
	return gr_warning("negative stdev: %.3g\n", d);
      if ( (pars[p].sd = d) > TINY)
	pars[p].noise = true;
      else
	pars[p].pos = pars[p].noise = false;
    }
    else
      return gr_warning("read error: expected a numerical argument: %s\n", 
			argv[i]);
  }
  else
    return 1;
  i++;
  if (argc>i) {
    if (abbrevArg(i, "positive", argc, argv))
      pars[p].pos = true;
    else
      return gr_warning("read error: %s\n",argv[i]);
  }
  else
    pars[p].pos = false;
  return 1;
}

int Nullcline(int argc, char *argv[]) {
  int i, n, ind;
  static boolean *toreport = NULL;

  /*  buffer keeping track of which var nullclines should be displayed  */
  if (toreport == NULL) {
    toreport = bcalloc(nvar, "'toreport' in Nullcline()()");
    /*  by default nullclines for all variables are displayed */
    for (i=0; i<nvar; i++)
      toreport[i] = true;
  }

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("nullcline variable variable"," ","Draw nullclines");
  if (eqArg(1, "status", argc, argv)) {
    printf("Status nullcline: showing nullclines for");
    for (n=0; n<nvar; n++)
      if (toreport[n])
	printf(" %s ", all_names[n]);
    printf("\n");
    return 1;
  }

  ind = 1;

  /*  more arguments indicate variables for which nullclines should be drawn */
  if (argc > ind) {
    for (n=0; n<nvar; n++)
      toreport[n] = false;
    for (i=ind; i<argc; i++)
      if ( (n = nameIndex(argv[i])) >= 0 && n <nvar)
	toreport[n] = true;
  }
  /*  draw nullclines for each variable in the toreport buffer */
  for (n=0; n<nvar; n++)
    if (toreport[n])
      nullcline(n);
  return 1;
}

#define n_options 17
struct _opt_table_ {
  char *name;
  boolean *opt;
} opt_table[n_options] = {
  {"adjust",    &adjust_opt},
  {"attained",  &attained_opt},
  {"axis",      &axis_opt},
  {"backward",  &backward_opt},
  {"calculator",&calcul_opt},
  {"closed",    &closed_opt},
  {"echo",      &echo_opt},
  {"erase",     &erase_opt},
  {"initial",   &initial_opt},
  {"keepvar",   &keepvar_opt},
  {"labels",    &label_opt},
  {"logfile",   &logfile_opt},
  {"map",       &map_opt},
  {"mark",      &mark_opt},
  {"points",    &points_opt},
  {"positive",  &positive_opt},
  {"truncate",  &trunc_opt}};


int Option(int argc,char *argv[]) {
  int i, j;

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) {
    gHelp("option option [off]"," ","Available options are:");
    for (i=0;  i < n_options; i++) {
      printf("%-12s",opt_table[i].name);
      if (i && i % 5 == 0) printf("\n");
    }
    printf("\n");
    return 1;
  }
  if (eqArg(1, "status", argc, argv)) {
    for (i=0; i<n_options; i++)
      printf("option %-10s  =\t%s\n", 
	     opt_table[i].name, *(opt_table[i].opt)?"true":"false");
    return 1;
  }

  for (i=1; i<argc; i++) {
    for (j=0; j<n_options; j++)
      if (eqf(argv[i], opt_table[j].name)) {
	if (eqArg(i+1, "off", argc, argv)) {
	  i++;
	  *(opt_table[j].opt) = false;
	}
	else
	  *(opt_table[j].opt) = true;
	
	break;
      }
    if (j == n_options)
      printf("unknown option: %s\n", argv[i]);
  }
  return 1;
}

int Output(int argc,char *argv[]) {
  int i, v;
  char flag = '+';

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("output variable expression parameter","output all variables",
		 "Select who is plotted in timeplots or printed during runs");
  if (eqArg(1, "status", argc, argv)) {
    printf("Status output: showing output for ");
    for (v=0; v<nnames; v++)
      if (outputbuf[v])
	printf("%s ",all_names[v]);
    printf("\n");
    return 1;
  }

  /*  set length of outputbuffer to zero  */
  /*  so we won't read from an unfilled buffer */
  rundatalength = 0;  /* this specifies the length of the databuffer  */
                      /* which is filled with proper data */

  /* check whether first argument contains a '+' or '-'
     in that case we will only update outputbuf
     otherwise we will completely reset it
  */
  
  if (argc == 1 || (argv[1][0] != '+' && argv[1][0] != '-'))
    for (v=0; v<nnames; v++)
      outputbuf[v] = false;

  /* go through all arguments; if first character is a +/- 
     set flag accordingly, increase pointer to that argument by one
     so that it now poiints to the remaining name of the variable
     (and break from loop if that name is empty, 
     ie there was a space between +/- and variable-name
     this parsing allows something like:
     output + var1 var2 -var2 var3 - var4 + var5
  */

  for (i=1; i<argc; i++) {
    if (('+' == argv[i][0]) || ('-' == argv[i][0])) {
      flag = argv[i][0];
      if ('\0' == *(++argv[i])) continue;
    }
    if ((v = nameIndex(argv[i])) >= 0 && v < nnames) {
      outputbuf[v] = (('-' == flag)?false:true);
    }
    else
      gr_warning("not a valid argument: %s\n", argv[i]);
  }

  return 1;
}

int Parameter(int argc,char *argv[]) {
  /*  check for file; if so write to file */
  /*  anycase; print all parameter + values to stdout */
  /*  if not initialized indicate this on stdout  */
  /*  but use default values when writing to file */

  FILE *par_file;
  int i;

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("parameters [filename]"," ",
		 "Print parameter values and/or save these to a file");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for parameter\n");

  /*  check for additional arguments, i.e. a filename to write output to */
  if (argc>1) {
    if ( NULL == (par_file = fopen(argv[1],"w")) )
      gr_warning("unable to open file %s\n", argv[1]);
    else {
      for (i=0; i<npar; i++) {
	fprintf(par_file,sformat,pars[i].name);
	fprintf(par_file,"=");
	fprintf(par_file,dformat,par_values[nexpr+i]);
	fprintf(par_file,"\n");
      }
      fclose(par_file);
    }
  }

  /*  print to stdout, check whether parameters are initialized */
  for (i=0; i<npar; i++)
    if (pars[i].init) {
      printf(sformat,pars[i].name);
      printf("=");
      printf(dformat,par_values[nexpr+i]);
      if (pars[i].noise) {
	printf(" (+/- ");
	printf(dformat, pars[i].sd);
	printf(")");
      }
      if (pars[i].pos)
	printf(" (pos)");
      printf("\n");
    }
    else {
      printf(sformat, pars[i].name);
      printf("not initialized\n");
    }
  return 1;
}

int Perturb(int argc,char *argv[]) {
  static int direction = 0;
  static double distance = 0.0001;

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("perturb vector [step size]","perturb 1 1e-4",
		 "Make a small step from the steady state on the eigenvector indicated");
  if (eqArg(1, "status", argc, argv))
    return printf("Status perturb: vector=%d distance=%.3g\n", 
		  direction+1, distance);

  /*  check whether eigen-vectors exist */
  if (eigvector == NULL)
    return gr_warning("call 'eigen' before calling 'perturb'\n");

  /*  set eigen-vecor along which to perturb */
  if (isnumeric(1, argc, argv)) {
    direction = atoi(argv[1]) - 1;
    if (direction < 0 || direction >= nvar)
      return gr_warning("not an eigenvector: %d\n", direction + 1);
  }
  else
    return gr_warning("read error\n");
  /*  check whether eigen-vector has a complex eigenvalue */
  if (eigcomp[eigsort[direction]])
    return gr_warning("Complex Eigenvalue: no 2D manifolds\n");
  /*  set distance over which to perturb */
  if (argc>2) {
    if (isnumeric(2, argc, argv))
      distance = atof(argv[2]);
    else
      return gr_warning("read error\n");
  }
    /*  do it */
  perturb(direction, distance);
  /*  print warning */
  if (backward_opt)
    gr_warning("Inset: Option BACK/TRUNC switched on!\n");
  /*  print state and derivatives */
  prider(istate);
  return 1;
}

int Plot(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("plot"," ","Draw the loaded data as a trajectory");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for plot\n");
  /*  plot content of rundata buffer */
  return plot();
}

int Readfile(int argc,char *argv[]) {
  /*  open, and read a file containing GRIND commands */
  /*  during the processing of this file control is NOT returned to calling */
  /*  program or to graphics driver */
  FILE *orig_input;          /*  keep track of current input stream */
  char *puline = NULL;       /*  store current uline */
  char *pcbegin,*pcend;      /*  store current cbegin and cend; */
  static char *filename = NULL;
  static boolean first = true;

  if (first) {
    first = false;
    filename = storeString(filename, modelname);
    filename = (char*)realloc(filename,(strlen(filename)+5)*sizeof(char));
    filename = strcat(filename, ".txt");
  }
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("readfile file","readfile model.txt",
		 "Read a file with GRIND commands");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for readfile\n");

  /*  store previous input stream */
  orig_input = cinput;
  if (argc>1) 
    filename = storeString(filename, argv[1]);

  if(!(cinput = fopen(filename,"r")))
    return gr_warning("error opening file: %s\n",filename);
  
  /*  store original command line */
  puline = storeString(puline,uline);
  /*  store the original pointers to begin and end of the current command */
  pcbegin = cbegin;
  pcend = cend;
  /*  read file and process it; on eof close it and reset input pointer */
  GrindMainLoop(false);
  fclose(cinput);
  /*  restore original input stream */
  cinput = orig_input;
  /*  restore original command line */
  strcpy(uline,puline);
  free(puline);
  /*  restore original pointers to begin and end of the original command */
  cbegin = pcbegin;
  cend = pcend;
  return 1;
}

int GR_RGB(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  int i;
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("rgb index Rvalue Gvalue Bvalue","rgb 0 0. 0. 0.",
		 "Set RGB values for entry 'index' in the colormap");
  if (eqArg(1, "status", argc, argv)) {
    printf("Current colormap:\n");
    printf(sformat, "Index:");
    printf(sformat, "R:");
    printf(sformat, "G:");
    printf(sformat, "B:");
    printf("\n");
    for (i=0; i<NCOLORS; i++)
      printf(" %-11d %-11.2g %-11.2g %-11.2g\n", i
	     , colormap[i][0], colormap[i][1], colormap[i][2]);
    return 1;
  }

  if (argc == 5) {
    i = atoi(argv[1]);
    colormap[i][0] = atof(argv[2]);
    colormap[i][1] = atof(argv[3]);
    colormap[i][2] = atof(argv[4]);
  }
  else
    return gr_warning("incorrect number of arguments\n");
  return 1;
}

int Run(int argc,char *argv[]) {
  /*  command to compute a trajectory  */
  extern double delta; /* step size of euler integrator */
  extern int ngrun;    /* counter for no. of graphical runs; determines color*/
  int v;               /* iterator */
  FILE* datafile = NULL;/* datafile */

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("run [file]"," ",
		 "Run the integration and/or save the results to a file");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for run\n");

  /*  check whether all variables and parameters are initialized */
  chckin();

  /* check whether a file should be opened for output */
  if (argc > 1) {
    if ((datafile = fopen(argv[1],"w")) == NULL)
      return gr_warning("unable to open file %s\n", argv[1]);
  }
  if (backward_opt) {
    forward_opt = false;
    printf("Derivatives run backwards!\n");
  }
  if (mode_opt == ALPHA) {
    if (itor == EULER)
      printf("Delta = %11.3g\n", delta);
    printf(sformat,"Time");
    for (v=0; v<nnames; v++)
      if (outputbuf[v])
	printf(sformat, all_names[v]);
    printf("\n");
  }

  /*  copy initial state to fstate; computing will be done using fstate */
  coparr(nvar, fstate, istate);
  /*  call integration routine */
  integr(datafile);
  ngrun++;

  printf(" End of %s\n",itors[itor]);
  if (datafile)
    fclose(datafile);
  forward_opt = true;

  return 1;
}

int Shade(int argc,char *argv[]) {
  static int ax = Z_AXIS;   /*  axis along which to shade */
  static int shint = 10;    /*  no. of crosses at which to compute nullclines */
  static int var = 0;       /*  variable to shade */
  boolean have_ax = false;  /*  temporary flag */

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("shade [x/y/z] variable [distance]","shade z var1 3",
		 "Fill a 3-dimensional nullcline plane");
  if (eqArg(1, "status", argc, argv))
    return printf("Status shade: axis=%s var [distance=%d]\n", 
		  (X_AXIS==ax ? "x" : (Y_AXIS==ax ? "y" : "z")), shint);

  if (argc>1) {
    var = nameIndex(argv[argc-1]);
    if (var>=0 && var<nvar) {
      /*  last arg is a variable */
      if (argc>2)
	/*  if there are 2 arguments an axis is specified; read it below */
	have_ax = true;
    }
    else if (isnumeric(argc-1, argc, argv)) {
      /*  last arg is numeric; it must be the distance */
      if ( (shint = atoi(argv[argc-1])) <= 0 )
	return gr_warning("negative distance in shade: %d\n", shint);
      /*  2nd last argument must be variable name */
      var = nameIndex(argv[argc-2]);
      if (var<0 || var>=nvar)
	return gr_warning("not a variable: %s\n", argv[argc-2]);
      if (argc>3)
	/*  if there are 3 arguments an axis is specified; read it below */
	have_ax = true;
    }
    else
      return gr_warning("read error\n");
  }
  if (have_ax) {
    /*  axis is always specified at 2nd argument */
    ax = Index("xyz", argv[1]);
    if (ax<X_AXIS || ax>Z_AXIS)
      return gr_warning("not an axis: %s\n", argv[1]);
  }
  shade(var, ax, shint);
  return 1;
}

int Takens(int argc,char *argv[]) {
  static int delta = 1;
  int orig_mode;
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("takens delta","takens 1","Make a Takens reconstruction");
  if (eqArg(1, "status", argc, argv))
    return printf("Status Takens: %d\n", delta);

  if (argc>1) {
    if (isnumeric(1, argc, argv))
      delta = atoi(argv[1]);
    else
      return gr_warning("read error\n");
  }

  orig_mode = mode_opt;
  mode_opt = DIM2;
  if (!checkgraphics())
    Graphics(1, NULL);
  if (erase_opt) Erase(0, NULL);
  DrawTakens(delta);
  mode_opt = orig_mode;
  return 1;
}

int ThreeD(int argc,char *argv[]) {
  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("3d"," ","Depict a 3-dimensional phase space");
  if (eqArg(1, "status", argc, argv))
    return printf("Status mode: %s\n", 
		  ((mode_opt == NONE) ? "none" :
		   ((mode_opt == ALPHA) ? "alpha-numeric" :
		    ((mode_opt == DIM2) ? "2d-graphics" :
		     ((mode_opt == DIM3) ? "3d_graphics" : "unknown mode")))));
  if (erase_opt) 
    Erase(argc, argv);
  mode_opt = DIM3;
  if (!checkgraphics())
    Graphics(1, NULL);
  Draw3Daxis();
  return 1;
}

int Timeplot(int argc,char *argv[]) {
  static boolean *toreport = NULL;
  int i, p;
  int tmp_mode;

  /*  buffer keeping track of which var/expr/par should be output  */
  /*  this is a subset for the set which is defined by command 'output' */
  /*  (ie which is in 'outputbuf') */
  if (toreport == NULL) {
    toreport = bcalloc(nnames, "'toreport' in Timeplot()");
    /*  by default the complete 'outputbuf' is used */
    for (i=0; i<nnames; i++)
      toreport[i] = true;
  }

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("timeplot [variable variable]","timeplot all variables",
		 "Depict a time plot of the last run"); 
  if (eqArg(1, "status", argc, argv)) {
    printf("Status timeplot; showing graphs for:\t");
    for (i=0; i<nnames; i++)
      if (toreport[i] && outputbuf[i])
	printf("%s ", all_names[i]);
    printf("\n");
    return 1;
  }

  if (argc>1) {
    /*  other arguments indicate which var's and/or expr's should be plotted */
    for (i=0; i<nnames; i++)
      toreport[i] = false;    
    for (i=1; i<argc; i++)
      if ((p=nameIndex(argv[i]))<0 || p>=(nnames))
	gr_warning("not a var/expr: %s\n",argv[i]);
      else
	toreport[p] = true;
  }
  if (0 == rundatalength)
    return gr_warning("no data is (yet) available for timeplot;\n");
  /*  make the timeplot */
  if (erase_opt) Erase(argc,argv);
  tmp_mode = mode_opt;
  mode_opt = DIM2;
  if (!checkgraphics())
    Graphics(1, NULL);
  drawTimeAxis();
  drawTimeSeries(toreport);
  mode_opt = tmp_mode;
  return 1;
}

int TwoD(int argc,char *argv[]) {

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("2d"," ","Depict a 2-dimensional phase space");
  if (eqArg(1, "status", argc, argv))
    return printf("current report mode = %s\n", 
		  ((mode_opt == NONE) ? "none" :
		   ((mode_opt == ALPHA) ? "alpha-numeric" :
		    ((mode_opt == DIM2) ? "2d-graphics" :
		     ((mode_opt == DIM3) ? "3d_graphics" : "unknown mode")))));
  if (erase_opt) Erase(argc, argv);
  mode_opt = DIM2;
  if (!checkgraphics())
    Graphics(1, NULL);
  Draw2Daxis();
  return 1;
}

int Triangle(int argc,char *argv[]) {
  static double *dpop = NULL; /*  derivatives vector */
  extern int tr_size;         /*  size of triangle */

  if (NULL == dpop) dpop = dcalloc(nvar, "'dpop' in Triangle()");

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("triangle size","triange 1.0",
		 "Indicate the vector field at the initial condition");
  if (eqArg(1, "status", argc, argv))
    return printf("Status triangle: size=%d\n", tr_size);

  if (argc > 1) {
    if (isnumeric(1, argc, argv))
      tr_size = atoi(argv[1]);
    else
      return gr_warning("read error\n");
  }
  fcn(istate, dpop);
  return triangle(istate, dpop);
}

int Value(int argc,char *argv[]) {
  /*  print variable, expression values, and parameter values to stdout */
  /*  if not initialized indicate this on stdout  */
  /*  argv can contain a list of variable/expression/parameter names */
  /*  which values should be printed. This list is stored as new default. */

  int i,p;
  static int *toreport=NULL; /*  store which var's/expr's/par's are printed */

  if (toreport == NULL) {
    toreport = bcalloc(nnames, "'toreport' in Value()");
    /*  by default all var's and expr's are printed; the par's are not */
    for (i=0; i<nvar+nexpr; i++)
      toreport[i]=1;
    for (i=nvar+nexpr; i<nnames; i++)
      toreport[i]=0;
  }

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("value variable expression parameter"," ",
		 "Print the current values");
  if (eqArg(1, "status", argc, argv)) {
    printf("currently showing values for:\n");
    for (i=0; i<nnames; i++)
      if (toreport[i])
	printf("%s\n", all_names[i]);
    return 1;
  }

  /*  additional arguments specify which var/expr/par's should be reported  */
  /*  remember this next time */
  if (argc>1) {
    for (i=0; i<nnames; i++)
      toreport[i]=0;
    for (i=1; i<argc; i++)
      if ((p=nameIndex(argv[i]))<0)
	gr_warning("not a var/expr/par: %s\n",argv[i]);
      else
	toreport[p]=1;
  }

  for (i=0; i<nvar; i++)
    if (toreport[i]) {
      printf(sformat,all_names[i]);
      printf("=");
      printf(dformat,istate[i]);
      printf("/");
      printf(dformat,fstate[i]);
      printf("\n");
    }
  for (i=nvar; i<nvar+nexpr; i++)
    if (toreport[i]) {
      printf(sformat,all_names[i]);
      printf("=");
      printf(dformat,par_values[i-nvar]);
      printf("\n");
    }
  for (i=nvar+nexpr; i<nnames; i++)
    if (toreport[i]) {
      if (pars[i-nvar-nexpr].init) {
	printf(sformat,all_names[i]);
	printf("=");
	printf(dformat,par_values[i-nvar]);
	printf("\n");
      }
      else {
	printf(sformat, all_names[i]);
	printf("=\tnot initialized\n");
      }
    }
  return 1;
}

int Vector(int argc,char *argv[]) {
  static int stps = 5;

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("vector nsteps","vector 5.0","Depict the vector field");
  if (eqArg(1, "status", argc, argv))
    return printf("Status vector: nsteps=%d\n", stps);

  if (argc>1) {
    if (isnumeric(1, argc, argv))
      stps = atoi(argv[1]);
    else
      return gr_warning("read error\n");
  }
  /*  draw the vector-field */
  return vectorfield(stps);
}

int Where(int argc,char *argv[]) {
  int cnt;                   /*  loops over argument list */
  int i;                     /*  iterator */
  static double *yt = NULL;  /*  state vector */
  FILE *outfile;             /*  output file */

  if (yt == NULL)
    /*  allocate an array of at least 3 elements so all 3 axes get a value */
    yt = dcalloc( (nvar<3 ? 3 : nvar), "'yt' in Where()");

  /*  check first whether help or status is asked for */
  if (eqArg(1, "help", argc, argv)) 
    return gHelp("where [initial] [file]"," ",
		 "Plot and print state and/or save in file");
  if (eqArg(1, "status", argc, argv))
    return printf("No status for where\n");

  /*  check whether all variables and parameters are initialized */
  chckin();
  /*  set cnt to 2nd argument; increase 'cnt' only an arg is used */
  cnt = 1;
  if (abbrevArg(cnt, "initial", argc, argv)) {
    cnt++;
    coparr(nvar, yt, istate);
  }
  else
    coparr(nvar, yt, fstate);

  if (argc>cnt) {
    if ( (outfile = fopen(argv[cnt], "w")) == NULL)
      return gr_warning("unable to open file %s\n", argv[cnt]);
    else {
      /*  print state to output-file */
      for (i=0; i<nvar; i++) {
	fprintf(outfile, sformat,all_names[i]);
	fprintf(outfile, "=");
	fprintf(outfile, dformat,yt[i]);
	fprintf(outfile, "\n");
      }
      fclose(outfile);
    }
  }
  /*  print values of 'yt' to terminal */
  prider(yt);
  /*  indicate position on graphics window */
  displayPosition(yt);
  return 1;
}

int ExecMacro(char *str) {
  extern char *cbegin;     /*  pointer to begin of command in 'cline' */
  extern char *cend;       /*  pointer to end of command in 'cline' */
  extern int margc;        /*  argc, determined in ParseArgs (not a handy way) */
  extern char *margv[];    /*  argv, idem */
  char *pcbegin = cbegin;  /*  for storing original 'cbegin' */
  char *pcend = cend;      /*  for storing original 'cend' */
  char c;                  /*  temporary */

  /*  set some pointers to the beginning of the string, 'str',  */
  /*  which holds the macro */
  cbegin = cend = str;
  /*  extract commands from str, and execute them, one command at a time */
  do {
    /*  extract next command from 'uline' */
    c = ParseArgs();
    /*  if there are commands, execute them */
    if (margc)
      Exec(margc,margv);
  }
  while (c != '\0');

  /*  reset command pointers to original values */
  cbegin = pcbegin;
  cend = pcend;

  return 1;
}

int Exec(int argc,char *argv[]) {
  int i;
  char *macrostr;

  /*  check for assignment */
  if(argc > 1 && 0 == strcmp(argv[1], "="))    return Assign(argc, argv);

  /*  check whether we are dealing with a macro */
  if ( (macrostr = ReturnData(LocateNode(macrolist, argv[0]))) )
    return ExecMacro(macrostr);
    

  /*  look for function corresponding to 'argv[0]'  */
  /*  in the hash-array 'command_table' */
  for (i=0; i<n_commands; i++)
    /*  'argv[0]' matches if at least two char's match */
    if (eqf(argv[0], command_table[i].name))
      /*  execute the command with arguments */
      return command_table[i].funcptr(argc, argv);

  /*  corresponding command is not found */
  gr_warning("error;command not recognized:\n%s\n",cline);
  return 0;
}
